#!/usr/bin/env python

import Image, ImageTk, Tkinter, sys, time, re, tkFileDialog
from math import exp, sin, cos, pi, sqrt
from optparse import OptionParser

class mandelbrotSet:
    def __init__(self):
        print 'created Mandelbrot set'

    def iter(self, z):
        c = z
        for i in range(maxIter):
            z = z*z + c
            if abs(z) > bailout:
                return i
        return maxIter

class Palette:
    def gauss(self, x, mu, sigma, amp=255.):
        return int(amp*exp(-(x-mu)*(x-mu)/(sigma*sigma)))

    def getColor(self, i):
        q = float(i)/float(maxIter+1)
        index = {'red':0, 'green':1, 'blue':2}
        rgb = [0,0,0]
        for curve in self.curves:
            rgb[index[curve[0]]] += self.gauss(q, curve[1], curve[2], curve[3])
        return (rgb[0],rgb[1],rgb[2])

    def setColors(self):
        self.color = []
        print 'setting up palette with', maxIter+1, 'colors'
        for i in range(maxIter+1):
            self.color.append(self.getColor(i))        

    def draw(self):
        for y in range(size):
            color = self.color[int(float(y)/float(size)*float(maxIter+1))]
            for x in range(self.imgWidth):
                self.image.putpixel((x, y), color)
        self.imgPal = ImageTk.PhotoImage(self.image)
        self.parent.canPal.create_image(self.imgWidth/2, size/2, image=self.imgPal)
        self.parent.canPal.update_idletasks()

    def load(self):
        self.parent.infoMsg('open palette')
        fileFormat = [('Palette file (ascii)','*.pal')]
        fileH = tkFileDialog.askopenfile(parent=self.parent.parent,
                                            filetypes=fileFormat,
                                            mode='r',
                                            title="Open palette")
        if fileH != None:
            self.curves = []
            for line in fileH.read().splitlines():
                print '\t'+line
                fields = line.split()
                amp = 255
                if (len(fields)==4): amp = int(fields[3])
                self.curves.append((fields[0], float(fields[1]), float(fields[2]), amp))
            fileH.close()
            self.reset()
        

    def reset(self):
        self.setColors()
        self.draw()

    def __init__(self, parent):
        self.imgWidth = 40
        self.parent = parent
        self.image = Image.new("RGB", (self.imgWidth, size), (0,0,0))
        self.imgPal = ImageTk.PhotoImage(self.image)
        self.curves = []
        self.curves.append(('red',   0.3 , 0.24, 255))
        self.curves.append(('green', 0.5 , 0.24, 255))
        self.curves.append(('blue',  0.72, 0.24, 255))
        self.setColors()
        
class zoomBoxTool:
    x0 = 0
    y0 = 0
    w  = 0
    center = complex(0,0)
    width = 3.0
    active = 0
    rectId = 0
    
    def activate(self, x, y):
        self.x0 = x
        self.y0 = y
        self.active = 1
        
    def draw(self, target, x1, y1):
        w = int(sqrt((self.x0-x1)*(self.x0-x1) + (self.y0-y1)*(self.y0-y1)))
        self.w = w
        target.labCursor.configure(text='width %s' % self.width)
        if (self.rectId!=0): target.canFract.delete(self.rectId)
        self.rectId = target.canFract.create_rectangle(self.x0-w, self.y0-w, self.x0+w, self.y0+w)
        
    def zoom(self, target):
        self.active = 0
        target.canFract.delete(self.rectId)
        self.rectId = 0
        if (self.width==0):
            target.infoMsg('Zoom aborted (zero width)')
            return
        target.labCursor.configure(text='')
        target.center = self.center
        target.width  = self.width

        w = self.w
        zoomIm = target.image.crop((self.x0-w, self.y0-w, self.x0+w, self.y0+w))
        target.image = zoomIm.resize((size, size))
        target.zoomin()


class Application:
    def plot(self):
        dx = self.width/self.size
        bl = self.center - complex(self.width/2, self.width/2)
        for x in range(self.size):
            for y in range(self.size):
                c = bl + complex(x*dx, y*dx)
                color = self.pal.color[self.fractal.iter(c)]
                self.image.putpixel((x, size-1-y), color)
            if (x%16==0):
                self.updateDisplay()
        self.updateDisplay()

    def plot1(self):
        dx = self.width/self.size
        bl = self.center - complex(self.width/2, self.width/2)
        for x in range(self.size):
            for y in range(self.size):
                if (x%2==0 and y%2==0): continue
                c = bl + complex(x*dx, y*dx)
                color = self.pal.color[self.fractal.iter(c)]
                self.image.putpixel((x, size-1-y), color)
            if (x%16==0):
                self.updateDisplay()
                if (self.stopRequest==1):
                    self.stopRequest=0
                    return
        self.updateDisplay()

    def plotN(self, n):
        dx = self.width/self.size
        bl = self.center - complex(self.width/2, self.width/2)
        for x in range(self.size):
            if (x%n==0):
                for y in range(self.size):
                    if (y%n==0):
                        if (n!=8 and x%(2*n)==0 and y%(2*n)==0): continue
                        c = bl + complex(x*dx, y*dx)
                        color = self.pal.color[self.fractal.iter(c)]
                        self.image.paste(color, (x, size-y-n-2, x+n, size-y))
                if (x%(16*n)==0): self.updateDisplay()
        self.updateDisplay()
        if (n>2) : self.plotN(n/2)
        if (n==2): self.plot1()

    def replot(self):
        self.parent.config(cursor="watch")
        self.frame += 1
        self.infoMsg('Processing image %d' % self.frame)
        timeStart = time.clock()
        self.plotN(8)
        
        self.parent.config(cursor="")
        timeStop = time.clock()
        timeDiff = round(timeStop-timeStart, 3)
        self.infoMsg('Image %i done in %.1f seconds' % (self.frame, timeDiff))

    def zoomin(self):
        self.updateDisplay()
        self.parent.config(cursor="watch")
        self.frame += 1
        self.infoMsg('Processing image %d' % self.frame)
        timeStart = time.clock()
        #self.plotN(8)
        self.plot()
        self.parent.config(cursor="")
        timeStop = time.clock()
        timeDiff = round(timeStop-timeStart, 3)
        self.infoMsg('Image %i done in %.1f seconds' % (self.frame, timeDiff))

    def infoMsg(self, msg):
        self.labInfo.configure(text=msg)
        print msg

    def interrupt(self):
        self.stopRequest = 1
        print 'request interruption'

    def save(self):
        fileFormat = [('Fractal file (ascii)','*.fract'), ('jpeg file', '*.jpg')]
        fileName = tkFileDialog.asksaveasfilename(parent=self.parent,
                                                  filetypes=fileFormat,
                                                  title="Save as")
        if len(fileName) > 0:
            fileH = open(fileName, 'w')
            if (re.search('.jpg', fileName)):
                self.image.save(fileH, 'JPEG')
            if (re.search('.fract', fileName)):
                fileBody = 'center: %s\nwidth: %f\nmaxIter: %i\n' % (self.center,
                                                                 self.width,
                                                                 maxIter)
                fileH.write(fileBody)
            self.infoMsg("saved to %s" % fileName)
            
    def openFile(self):
        self.infoMsg('file open')
        fileFormat = [('Fractal file (ascii)','*.fract')]
        fileH = tkFileDialog.askopenfile(parent=self.parent,
                                            filetypes=fileFormat,
                                            mode='rb',
                                            title="Open file")
        if fileH != None:
            global maxIter
            for line in fileH.read().splitlines():
                print '\t'+line
                fields = line.split(': ')
                if (fields[0]=='maxIter'):
                    maxIter = int(fields[1])
                    self.entMax.delete(0, 'end')
                    self.entMax.insert(0, str(maxIter))
                if (fields[0]=='center'):
                    self.center = complex(fields[1].strip('()'))
                if (fields[0]=='width'):  self.width = float(fields[1])
            fileH.close()
            self.pal.reset()
            self.replot()

    def reset(self, makePlot=True):
        self.center = complex(-0.65, 0)
        self.width = 2.8
        self.frame = 0
        if (makePlot): self.replot()
        
    def evtMouseDown(self, event):
        bl = self.center - complex(self.width/2, self.width/2)
        z = bl + complex(event.x*self.width/size, (size-1-event.y)*self.width/size)
        print 'recenter at %s' % z
        self.infoMsg('zoom')
        self.zoomBox.activate(event.x, event.y)
        self.zoomBox.center = z

    def evtMouseMove(self, event):
        bl = self.center - complex(self.width/2, self.width/2)
        z = bl + complex(event.x*self.width/size, (size-1-event.y)*self.width/size)
        if (self.zoomBox.active==1):
            self.zoomBox.width = 2.0*abs(self.zoomBox.center - z)
            self.zoomBox.draw(self, event.x, event.y)
        else:
            self.labCursor.configure(text='%s' % z)
        
    def evtMouseUp(self, event):
        bl = self.center - complex(self.width/2, self.width/2)
        z = bl + complex(event.x*self.width/size, (size-1-event.y)*self.width/size)
        self.zoomBox.width = 2.0*abs(self.zoomBox.center - z)
        self.infoMsg('zoom with width '+str(self.zoomBox.width))
        self.zoomBox.zoom(self)

    def evtLeaveCanvas(self, event):
        self.labCursor.configure(text='')
        
    def setMaxIter(self):
        newMax  = self.entMax.get()
        #newBail = self.entBail.get()
        if (newMax.isdigit()): 
            global maxIter, bailout
            maxIter = int(newMax)
            #bailout = float(newBail)
            #self.infoMsg('maxIter: %i, bailout: %f' % (maxIter, bailout))
            self.infoMsg('maxIter: %i' % (maxIter))
            self.pal.reset()
        else:
            self.infoMsg('invalid integer: '+str(newMax))
            self.entMax.delete(0, 'end')
            self.entMax.insert(0, str(maxIter))
        self.btnPlot.focus_set()
            
    def __init__(self, parent, imageSize):
        self.parent = parent
        self.fractal = mandelbrotSet()
        self.pal = Palette(self)
        self.zoomBox = zoomBoxTool()
        self.size = imageSize
        self.image = Image.new("RGB", (imageSize,imageSize), self.pal.color[0])
        self.stopRequest=0
        
        self.createWidgets()
        self.pal.draw()
        self.reset(False)

    def createWidgets(self):
        self.imgFract = ImageTk.PhotoImage(self.image)
        
        self.frmTop = Tkinter.Frame(self.parent)
        self.frmTop.pack(anchor="w")
        self.frmCan = Tkinter.Frame(self.parent)
        self.frmCan.pack(anchor="w", in_=self.frmTop, side="left")
        self.frmPal = Tkinter.Frame(self.parent)
        self.frmPal.pack(anchor="w", in_=self.frmTop, side="left")

        self.canFract = Tkinter.Canvas(self.parent, width=size, height=size)        
        self.canFract.bind("<Button-1>", self.evtMouseDown)
        self.canFract.bind("<Motion>", self.evtMouseMove)
        self.canFract.bind("<ButtonRelease-1>", self.evtMouseUp)
        self.canFract.bind("<KeyPress-C>", self.interrupt)
        self.canFract.bind("<Leave>", self.evtLeaveCanvas)
        self.canFract.pack(in_=self.frmCan)

        self.canPal = Tkinter.Canvas(self.parent, width=self.pal.imgWidth, height=size)        
        self.canPal.pack(in_=self.frmPal)
        
        self.frm1 = Tkinter.Frame(self.parent)
        self.frm1.pack(anchor="w", fill='x')
        self.frm2 = Tkinter.Frame(self.parent)
        self.frm2.pack(anchor="w", fill='x')
        self.frm3 = Tkinter.Frame(self.parent)
        self.frm3.pack()
        self.frm4 = Tkinter.Frame(self.parent)
        self.frm4.pack(anchor="s", in_=self.frmTop, side="left")

        #self.btnStop = Tkinter.Button(self.parent, text="Cancel", command=self.interrupt)
        #self.btnStop.pack(side="left", in_=self.frm1)

        self.labInfo = Tkinter.Label(self.parent, text="Initializing...")
        self.labInfo.pack(side="right", in_=self.frm1)

        self.entMax = Tkinter.Entry(self.parent, width=4)
        self.entMax.insert(0, str(maxIter))
        self.entMax.pack(side="left", in_=self.frm1)

        self.labIter = Tkinter.Label(self.parent, text='iterations')
        self.labIter.pack(side='left', in_=self.frm1)

        #self.entBail = Tkinter.Entry(self.parent, width=4)
        #self.entBail.insert(0, str(bailout))
        #self.entBail.pack(side="left", in_=self.frm2)

        self.btnIter = Tkinter.Button(self.parent, text='Apply', command=self.setMaxIter)
        self.btnIter.pack(side='left', in_=self.frm1)

        self.labCursor = Tkinter.Label(self.parent, text="", relief='sunken')
        self.labCursor.pack(in_=self.frm2, fill='x')

        self.btnPlot = Tkinter.Button(self.parent, text="Plot", command=self.replot)
        self.btnPlot.pack(in_=self.frm4, fill='x')

        self.btnReset = Tkinter.Button(self.parent, text="Reset", command=self.reset)
        self.btnReset.pack(in_=self.frm4, fill='x')

        self.btnSave = Tkinter.Button(self.parent, text="Save", command=self.save)
        self.btnSave.pack(in_=self.frm4, fill='x')

        self.btnOpen = Tkinter.Button(self.parent, text="Open", command=self.openFile)
        self.btnOpen.pack(in_=self.frm4, fill='x')

        self.btnPal  = Tkinter.Button(self.parent, text="Palette", command=self.pal.load)
        self.btnPal.pack(in_=self.frm4, fill='x')

        self.btnQuit = Tkinter.Button(self.parent, text="Exit", command=self.parent.quit)
        self.btnQuit.pack(in_=self.frm4, fill='x')


    def updateDisplay(self):
        self.imgFract = ImageTk.PhotoImage(self.image)
        self.canFract.create_image(size/2, size/2, image=self.imgFract)
        self.canFract.update_idletasks()


# options
parser = OptionParser()
parser.add_option("-s", "--size", type="int", dest="size",
                  help="size of image in pixels", default=600)
parser.add_option("-m", "--max", type="int", dest="maxIter",
                  help="maximum number of iterations", default=250)
(options, args) = parser.parse_args()


bailout = 2.0
maxIter = options.maxIter
size = options.size

App = Tkinter.Frame()
App.pack()
App.master.title("Mandelbrot Set")
im = Application(App, size)

App.mainloop()

